﻿Option Strict On
Option Explicit On 
'☆ このソースコードの解読は､ 許諾されています。
'#SkipAnalysis このモジュールには、MANDALA フックメソッドが含まれていないので、MANDALA による解析処理の対象から外します。
Imports AppliTech.WorkFrame
Imports System '基本 Imports
Imports VB = Microsoft.VisualBasic

Namespace AppliFrame.DBMS
    Public MustInherit Class TextDBMS

        'TextDBMS (テキストファイルを用いた SQL 風のアクセスを可能にする DBMS)

        '  次の５機能をサポートしています。
        '    コネクト処理
        '    シークエンシャルアクセス
        '    SQL 風のアクセス
        '    トランザクション制御
        '    右詰左詰処理

        '  以下にメソッド一覧を示しますが、これらを用いるにはこのサブクラスを宣言して、それをコンストラクトすることが必要です。

        '★ ｢コネクト処理｣ に関するメソッド
        '
        '      Connect
        '      Disconnect

        '★ ｢シークエンシャルアクセス｣ に関するメソッド
        '
        '      SetCursorForAllRecords
        '      ReadSequential

        '★ ｢SQL 風のアクセス｣ に関するメソッド
        '
        '    Select 風のアクセスメソッド
        '      SelectEqual
        '      SelectNext
        '      SelectPrev
        '      SelectFirst
        '      SelectLast
        '
        '    Insert 風のアクセスメソッド
        '      InsertRecord
        '
        '    Update 風のアクセスメソッド
        '      UpdateRecord
        '
        '    Delete 風のアクセスメソッド
        '      DeleteRecord

        '★ ｢トランザクション制御｣ に関するメソッド
        '
        '      TranStart
        '      TranCommit
        '      TranAbort

        '★ ｢右詰左詰処理｣ に関するメソッド
        '
        '      AlignRight
        '      AlignLeft

#If CentralOrOp OrElse Not LocalOrOp Then 'ワンピースまたはセントラルピース用 (つまりローカルピースを除く)

        Protected Record As String 'Record
        Private RecordLength As Integer 'レコード長 (バイト数)
        Public RecordNumber As Integer 'レコード番号
        '
        Private Filename As String 'ファイル名 (フルパス)

        'コンストラクタ
        Public Sub New(ByVal pm_Filename As String, ByVal pm_RecordLength As Integer)
            Me.Filename = pm_Filename
            Me.RecordLength = pm_RecordLength
        End Sub 'New

        'このサブクラスで宣言することが必要なメソッド群
        Protected MustOverride Sub Serialize()
        Protected MustOverride Sub DeSerialize()
        Public MustOverride Function GetUniqueKey() As String

        Public Overridable Function GetBodyDetailCount() As Integer
            Return 0 '----------
        End Function 'GetBodyDetailCount

        Public Overridable Function GetEBodyDetailCount() As Integer
            Return 0 '----------
        End Function 'GetEBodyDetailCount

        '★ ｢コネクト処理｣ に関するメソッドなど =====================================================================================

        Private ConnectedFlag As Boolean 'Connect 済み表示
        Private ReadWriteFlag As Boolean 'Connect が ReadWrite モード

        'Connect します。
        Public Function Connect() As Integer
            If ConnectedFlag Then
                Throw New WorkFrameException("｢重複コネクト (Connect)｣")
            Else
                ConnectedFlag = True
                ReadWriteFlag = False
                '
                'Connect 連鎖に付け加える。
                Me.nextTable = ConnectedReadWriteFirst
                ConnectedReadWriteFirst = Me
                '
                Return ConnectHelper() '----------
            End If
        End Function 'Connect

        'Disconnect します。
        Public Sub Disconnect()
            If ConnectedFlag Then
                'Connect 連鎖から取り除く。
                If ConnectedReadWriteFirst Is Me Then
                    ConnectedReadWriteFirst = Me.nextTable
                Else
                    Dim wk_Table As TextDBMS = ConnectedReadWriteFirst
                    While Not (wk_Table.nextTable Is Me)
                        wk_Table = wk_Table.nextTable
                    End While
                    wk_Table.nextTable = Me.nextTable
                End If
                '
                ClearRow()
                ReadWriteFlag = False
            Else
                Throw New WorkFrameException("｢未コネクト (Disconnect)｣")
            End If
        End Sub 'Disconnect

        '★ ｢シークエンシャルアクセス｣ に関するメソッドなど =========================================================================

        Private Cursor As RowData '順次読み込み用の Cursor 

        'Cursor を初期化して順次読み込み (ReadSequential) の準備をします。
        Public Sub SetCursorForAllRecords()
            Cursor = Nothing
        End Sub 'SetCursorForAllRecords

        'Cursor を用いてレコードを順次読み込みます。
        Public Function ReadSequential() As Integer
            '次の Cursor の位置を求める。
            If Cursor Is Nothing Then
                If RowFirst Is Nothing Then
                    Return -1 'EOF
                Else
                    Cursor = RowFirst
                End If
            Else
                If Cursor Is RowLast Then
                    Return -1 'EOF
                Else
                    Cursor = Cursor.nextRow
                End If
            End If
            '
            Record = Cursor.RowElements
            DeSerialize()
            Return 0 '----------
        End Function 'ReadSequential

        '★ ｢SQL 風のアクセス｣ に関するメソッド =====================================================================================

        '指定のキーと一致するレコードの各データをそれぞれのカラムに設定します。
        Public Function SelectEqual(ByVal pm_KeyString As String) As Integer
            Dim wk_row As RowData
            If RowFirst Is Nothing Then
                Return 100 'レコードがない。
            Else
                wk_row = RowFirst
            End If
            Dim wk_RecordNumber As Integer = 0
            '
            While Not (wk_row Is Nothing)
                wk_RecordNumber += 1
                Record = wk_row.RowElements
                If GetUniqueKey() = pm_KeyString Then
                    DeSerialize()
                    RecordNumber = wk_RecordNumber
                    Return 0 '----------
                End If
                wk_row = wk_row.nextRow '次の Row を取り出す
            End While
            '
            Return 101 '条件に合う (指定のキーと一致する) レコードがみつからない。
        End Function 'SelectEqual

        '指定のキーより大きい (ものの中で最小の) キーをもつレコードの各データをそれぞれのカラムに設定します。
        Public Function SelectNext(ByVal pm_KeyString As String) As Integer
            Dim wk_row As RowData
            If RowFirst Is Nothing Then
                Return 100 'レコードがない。
            Else
                wk_row = RowFirst
            End If
            Dim wk_RecordNumber As Integer = 0
            '
            While Not (wk_row Is Nothing)
                wk_RecordNumber += 1
                Record = wk_row.RowElements
                If GetUniqueKey().CompareTo(pm_KeyString) > 0 Then
                    DeSerialize()
                    RecordNumber = wk_RecordNumber
                    Return 0 '----------
                End If
                wk_row = wk_row.nextRow '次の Row を取り出す
            End While
            '
            Return 102 '条件に合う (指定のキーより大きい) レコードがみつからない。
        End Function 'SelectNext

        '指定のキーより小きい (ものの中で最大の) キーをもつレコードの各データをそれぞれのカラムに設定します。
        Public Function SelectPrev(ByVal pm_KeyString As String) As Integer
            'この場合は、特別に逆順にさがす。
            Dim wk_row As RowData
            If RowLast Is Nothing Then
                Return 100 'レコードがない。
            Else
                wk_row = RowLast
            End If
            Dim wk_RecordNumber As Integer = RecordCount
            '
            While Not (wk_row Is Nothing)
                Record = wk_row.RowElements
                If GetUniqueKey().CompareTo(pm_KeyString) < 0 Then
                    DeSerialize()
                    RecordNumber = wk_RecordNumber
                    Return 0 '----------
                End If
                wk_RecordNumber -= 1
                wk_row = wk_row.prevRow '一つ前の Row を取り出す
            End While
            '
            Return 103 '条件に合う (指定のキーより小さい) レコードがみつからない。
        End Function 'SelectPrev

        '先頭レコードの各データをそれぞれのカラムに設定します。
        Public Function SelectFirst() As Integer
            If RowFirst Is Nothing Then
                Return 100 'レコードがない。
            Else
                Record = RowFirst.RowElements
                DeSerialize()
                RecordNumber = 1
                Return 0 '----------
            End If
        End Function 'SelectFirst

        '最終レコードの各データをそれぞれのカラムに設定します。
        Public Function SelectLast() As Integer
            If RowLast Is Nothing Then
                Return 100 'レコードがない。
            Else
                Record = RowLast.RowElements
                DeSerialize()
                RecordNumber = RecordCount
                Return 0 '----------
            End If
        End Function 'SelectLast

        'レコードの各カラム (キーも含む) に値を設定してから、InsertRecord を呼び出してください。
        'すると、そのレコードが挿入されます。
        'ただし、InsertRecord を呼び出す前に、TranStart を呼び出し、最後に TranCommit を呼び出すことが必要です。
        'なお、実際の挿入処理は、TranCommit によってなされます。
        Public Function InsertRecord() As Integer
            If TranMode Then
                Serialize()
                Dim wk_NewRow As RowData = New RowData(Record)
                Dim sg_KeyString As String = GetUniqueKey()
                '
                Dim wk_row As RowData = RowFirst
                While Not (wk_row Is Nothing)
                    Record = wk_row.RowElements
                    If GetUniqueKey() = sg_KeyString Then
                        Return 112 '指定のキーのレコードが存在する。
                    ElseIf GetUniqueKey().CompareTo(sg_KeyString) > 0 Then
                        InsertPrevRowHelper(wk_row, wk_NewRow)
                        ReadWriteFlag = True
                        Return 0 '----------
                    End If
                    wk_row = wk_row.nextRow '次の Row を取り出す
                End While
                '
                AppendRowHelper(wk_NewRow)
                ReadWriteFlag = True
                Return 0 '----------
            Else
                Throw New WorkFrameException("｢まだ TranStart が呼び出されていないか、ReadOnly として Connect されている (InsertRecord)｣")
            End If
        End Function 'InsertRecord

        'レコードの各カラムに値を設定してから、UpdateRecord を呼び出してください。
        'すると、パラメタで指定したキーをもつレコードの内容が更新されます。
        'ただし、UpdateRecord を呼び出す前に、TranStart を呼び出し、最後に TranCommit を呼び出すことが必要です。
        'なお、実際の更新処理は、TranCommit によってなされます。
        Public Function UpdateRecord(ByVal pm_KeyString As String) As Integer
            If TranMode Then
                Dim wk_row As RowData
                If RowFirst Is Nothing Then
                    Return 100 'レコードがない。
                Else
                    wk_row = RowFirst
                End If
                '
                Serialize()
                If Not GetUniqueKey().Equals(pm_KeyString) Then
                    Return 111 '指定のキーとレコードのキーが一致しない。
                End If
                '
                While Not (wk_row Is Nothing)
                    Record = wk_row.RowElements
                    If GetUniqueKey() = pm_KeyString Then
                        Serialize() '目指す Record を形成するために必要。
                        wk_row.RowElements = Record
                        ReadWriteFlag = True
                        Return 0 '----------
                    End If
                    wk_row = wk_row.nextRow '次の Row を取り出す
                End While
                Return 101 '条件に合う (指定のキーと一致する) レコードがみつからない。
            Else
                Throw New WorkFrameException("｢まだ TranStart が呼び出されていないか、ReadOnly として Connect されている (UpdateRecord)｣")
            End If
        End Function 'UpdateRecord

        'キーを指定して、DeleteRecord を呼び出してください。
        'すると、パラメタで指定したキーをもつレコードが削除されます。
        'ただし、DeleteRecord を呼び出す前に、TranStart を呼び出し、最後に TranCommit を呼び出すことが必要です。
        'なお、実際の削除処理は、TranCommit によってなされます。
        Public Function DeleteRecord(ByVal pm_KeyString As String) As Integer
            If TranMode Then
                Dim wk_row As RowData
                If RowFirst Is Nothing Then
                    Return 100 'レコードがない。
                Else
                    wk_row = RowFirst
                End If
                '
                While Not (wk_row Is Nothing)
                    Record = wk_row.RowElements
                    If GetUniqueKey() = pm_KeyString Then
                        DeleteRowHelper(wk_row)
                        ReadWriteFlag = True
                        Return 0 '----------
                    End If
                    wk_row = wk_row.nextRow '次の Row を取り出す
                End While
                Return 101 '条件に合う (指定のキーと一致する) レコードがみつからない。
            Else
                Throw New WorkFrameException("｢まだ TranStart が呼び出されていないか、ReadOnly として Connect されている (DeleteRecord)｣")
            End If
        End Function 'DeleteRecord

        '★ ｢トランザクション制御｣ に関するメソッドなど =============================================================================

        Private TranMode As Boolean '個々の TextDBMS ごとの TranStarted 状態表示。
        Private Shared TranStarted As Boolean = False 'TranStarted 状態表示 (Shared)。

        Private Shared ConnectedReadWriteFirst As TextDBMS = Nothing 'Connect 連鎖の始点 (Shared)。
        Private nextTable As TextDBMS '次の Connect 連鎖。

        'トランザクションを開始させます。
        'UpdateRecord, InsertRecord, DeleteRecord の前に呼び出すことが必要。
        Public Shared Sub TranStart()
            If TranStarted Then
                Throw New WorkFrameException("｢すでに TranStart が呼び出されている (TranStart)｣")
            Else
                Dim wk_Table As TextDBMS = ConnectedReadWriteFirst
                While Not (wk_Table Is Nothing)
                    wk_Table.TranMode = True
                    wk_Table = wk_Table.nextTable '次の ReadWrite TextRDBMS を取り出す
                End While
                '
                TranStarted = True
            End If
        End Sub 'TranStart

        'トランザクションを正常終了させます。
        'UpdateRecord, InsertRecord, DeleteRecord の後に呼び出すことが必要。
        Public Shared Sub TranCommit()
            If TranStarted Then
                Dim wk_Table As TextDBMS = ConnectedReadWriteFirst
                While Not (wk_Table Is Nothing)
                    If wk_Table.ReadWriteFlag Then
                        wk_Table.ReadWriteFlag = False
                        '
                        wk_Table.TranMode = False
                        If wk_Table.CommitHelper() <> 0 Then
                            Throw New WorkFrameException("｢データの一貫性が損なわれた (TranCommit)｣")
                        End If
                    End If
                    wk_Table = wk_Table.nextTable '次の ReadWrite TextRDBMS を取り出す
                End While
                '
                TranStarted = False
            Else
                Throw New WorkFrameException("｢まだ TranStart が呼び出されていない (TranCommit)｣")
            End If
        End Sub 'TranCommit

        'トランザクションを異常終了させます。
        'UpdateRecord, InsertRecord, DeleteRecord の後に呼び出すことが必要。
        Public Shared Sub TranAbort()
            If TranStarted Then
                Dim wk_Table As TextDBMS = ConnectedReadWriteFirst
                While Not (wk_Table Is Nothing)
                    If wk_Table.ReadWriteFlag Then
                        wk_Table.ReadWriteFlag = False
                        '
                        wk_Table.TranMode = False
                        If wk_Table.ConnectHelper() <> 0 Then
                            Throw New WorkFrameException("｢データの一貫性が損なわれた (TranAbort)｣")
                        End If
                    End If
                    wk_Table = wk_Table.nextTable '次の ReadWrite TextRDBMS を取り出す
                End While
                '
                TranStarted = False
            Else
                Throw New WorkFrameException("｢まだ TranStart が呼び出されていない (TranAbort)｣")
            End If
        End Sub 'TranAbort

        '★ ｢右詰左詰処理｣ に関するメソッド =========================================================================================

        'カラムの内容を右詰にします。
        Public Shared Function AlignRight(ByVal pm_String As String, ByVal pm_Length As Integer) As String
            Return FormBase.RightWid(VB.Strings.Space(pm_Length) + pm_String, pm_Length) '----------
        End Function 'AlignRight

        'カラムの内容を左詰にします。
        Public Shared Function AlignLeft(ByVal pm_String As String, ByVal pm_Length As Integer) As String
            Return FormBase.LeftWid(pm_String + VB.Strings.Space(pm_Length), pm_Length) '----------
        End Function 'AlignLeft

#Region "TextDBMS の各種 Helper メソッド"

        'Connect, TranAbort 用の Helper メソッド (Private)。
        Private Function ConnectHelper() As Integer
            ClearRow()
            '
            Dim wk_FileID As OpenedInputFile = FileInputOpen(Filename, System.IO.FileMode.Open)
            If wk_FileID Is Nothing Then
                'オープンエラー。
                Return 1 '----------
            Else
                Dim Ex_KeyString As String = Nothing
                While Not EOF(wk_FileID)
                    Record = LineInput(wk_FileID)
                    If FormBase.LenWid(Record) <> RecordLength Then
                        FileInputClose(wk_FileID)
                        Throw New WorkFrameException("｢データのレコード長に関する誤り (ConnectHelper)｣")
                    End If
                    '
                    Dim sg_KeyString As String = GetUniqueKey()
                    If Ex_KeyString Is Nothing Then
                    ElseIf sg_KeyString.Equals(Ex_KeyString) Then
                        FileInputClose(wk_FileID)
                        Throw New WorkFrameException("｢データのユニークキーの値が重複 (ConnectHelper)｣")
                    ElseIf sg_KeyString.CompareTo(Ex_KeyString) < 0 Then
                        FileInputClose(wk_FileID)
                        Throw New WorkFrameException("｢データがキーの順にソートされていない (ConnectHelper)｣")
                    End If
                    Ex_KeyString = sg_KeyString
                    '
                    Dim wk_row As RowData = New RowData(Record)
                    AppendRowHelper(wk_row)
                End While
                '
                '正常終了。
                FileInputClose(wk_FileID)
                Return 0 '----------
            End If
        End Function 'ConnectHelper

        'TranCommit 用の Helper メソッド (Private)。
        Private Function CommitHelper() As Integer
            Dim wk_FileID As OpenedOutputFile = FileOutputOpen(Filename, System.IO.FileMode.Create)
            If wk_FileID Is Nothing Then
                Return -1 '----------
            Else
                Dim wk_row As RowData = RowFirst
                While Not (wk_row Is Nothing)
                    Record = wk_row.RowElements
                    PrintLine(wk_FileID, Record)
                    wk_row = wk_row.nextRow '次の Row を取り出す
                End While
                '
                FileOutputClose(wk_FileID)
                '
                Return 0 '----------
            End If
        End Function 'CommitHelper

#End Region 'TextDBMS の各種 Helper メソッド

#Region "TextDBMS が使用する Text File System"

        Private Shared Encoding As System.Text.Encoding = System.Text.Encoding.GetEncoding(932) '"shift_jis", "iso-2022-jp"
        Private Shared ExceptionOfFileOpen As System.Exception

        Private Shared Function FileInputOpen(ByVal pm_FileName As String, ByVal pm_FileMode As System.IO.FileMode) As OpenedInputFile
            Dim if_OpenedInputFile As OpenedInputFile = New OpenedInputFile(pm_FileName, pm_FileMode)
            Try
                if_OpenedInputFile.OpenedStream = System.IO.File.Open(pm_FileName, pm_FileMode, System.IO.FileAccess.Read, System.IO.FileShare.Read)
                if_OpenedInputFile.OpenedStreamReader = New System.IO.StreamReader(if_OpenedInputFile.OpenedStream, Encoding, True)
                '
                if_OpenedInputFile.nextMyObj = OpenedInputFileFirst
                OpenedInputFileFirst = if_OpenedInputFile
                Return if_OpenedInputFile '----------
            Catch ex As System.Exception
                ExceptionOfFileOpen = ex
                Return Nothing '----------
            End Try
        End Function 'FileInputOpen

        Private Overloads Shared Function FileOutputOpen(ByVal pm_FileName As String) As OpenedOutputFile
            Return FileOutputOpen(pm_FileName, System.IO.FileMode.Create, Encoding) '----------
        End Function 'FileOutputOpen
        Private Overloads Shared Function FileOutputOpen(ByVal pm_FileName As String, ByVal pm_FileMode As System.IO.FileMode) As OpenedOutputFile
            'System.IO.FileMode.Append の場合であり、Encoding は、以前のものを継承するので、意味がない。
            Return FileOutputOpen(pm_FileName, pm_FileMode, Encoding) '----------
        End Function 'FileOutputOpen
        Private Overloads Shared Function FileOutputOpen(ByVal pm_FileName As String, ByVal pm_FileMode As System.IO.FileMode, ByVal pm_Encoding As System.Text.Encoding) As OpenedOutputFile
            Dim of_OpenedOutputFile As OpenedOutputFile = New OpenedOutputFile(pm_FileName, pm_FileMode)
            Try
                of_OpenedOutputFile.OpenedStream = System.IO.File.Open(pm_FileName, pm_FileMode, System.IO.FileAccess.Write, System.IO.FileShare.Write)
                If pm_FileMode.Equals(System.IO.FileMode.Append) Then
                    'System.IO.FileMode.Append の場合、Encoding は、以前のものを継承する。
                    of_OpenedOutputFile.OpenedStreamWriter = New System.IO.StreamWriter(of_OpenedOutputFile.OpenedStream, pm_Encoding) '----------
                Else
                    of_OpenedOutputFile.OpenedStreamWriter = New System.IO.StreamWriter(of_OpenedOutputFile.OpenedStream, pm_Encoding) '----------
                End If
                '
                of_OpenedOutputFile.nextMyObj = OpenedOutputFileFirst
                OpenedOutputFileFirst = of_OpenedOutputFile
                Return of_OpenedOutputFile '----------
            Catch ex As System.Exception
                ExceptionOfFileOpen = ex
                Return Nothing '----------
            End Try
        End Function 'FileOutputOpen

        Private Shared Sub FileInputClose(ByVal pm_OpenedInputFile As OpenedInputFile)
            pm_OpenedInputFile.OpenedStreamReader.Close()
            pm_OpenedInputFile.OpenedStream.Close()
            '
            'Open 連鎖から取り除く。
            If OpenedInputFileFirst Is pm_OpenedInputFile Then
                OpenedInputFileFirst = pm_OpenedInputFile.nextMyObj
            Else
                Dim if_UnclodedFile As OpenedInputFile = OpenedInputFileFirst
                While Not (if_UnclodedFile.nextMyObj Is pm_OpenedInputFile)
                    if_UnclodedFile = if_UnclodedFile.nextMyObj
                End While
                if_UnclodedFile.nextMyObj = pm_OpenedInputFile.nextMyObj
            End If
        End Sub 'FileInputClose

        Private Shared Sub FileOutputClose(ByVal pm_OpenedOutputFile As OpenedOutputFile)
            pm_OpenedOutputFile.OpenedStreamWriter.Close()
            pm_OpenedOutputFile.OpenedStream.Close()
            '
            'Open 連鎖から取り除く。
            If OpenedOutputFileFirst Is pm_OpenedOutputFile Then
                OpenedOutputFileFirst = pm_OpenedOutputFile.nextMyObj
            Else
                Dim of_UnclodedFile As OpenedOutputFile = OpenedOutputFileFirst
                While Not (of_UnclodedFile.nextMyObj Is pm_OpenedOutputFile)
                    of_UnclodedFile = of_UnclodedFile.nextMyObj
                End While
                of_UnclodedFile.nextMyObj = pm_OpenedOutputFile.nextMyObj
            End If
        End Sub 'FileOutputClose

        Private Shared OpenedInputFileFirst As OpenedInputFile = Nothing
        Private Shared OpenedOutputFileFirst As OpenedOutputFile = Nothing

        Private Shared Function EOF(ByVal pm_OpenedInputFile As OpenedInputFile) As Boolean
            If pm_OpenedInputFile.OpenedStream Is Nothing OrElse pm_OpenedInputFile.OpenedStream.Length = 0 Then
                Return True '----------
            ElseIf pm_OpenedInputFile.OpenedStreamLastLines Is Nothing Then
                Return False '----------
            ElseIf pm_OpenedInputFile.OpenedStreamLastLines.Length > 0 Then
                Return False '----------
            Else
                Return True '----------
            End If
        End Function 'EOF

        Private Shared Function LineInput(ByVal pm_OpenedInputFile As OpenedInputFile) As String
            If pm_OpenedInputFile.OpenedStreamLastLines Is Nothing Then
                Dim sg_ReadLine As String = pm_OpenedInputFile.OpenedStreamReader.ReadLine()
                If pm_OpenedInputFile.OpenedStream.Position >= pm_OpenedInputFile.OpenedStream.Length Then
                    Dim sg_LastLine As String = pm_OpenedInputFile.OpenedStreamReader.ReadToEnd()
                    If sg_LastLine.EndsWith(VB.Constants.vbCrLf) Then
                        pm_OpenedInputFile.OpenedStreamLastLines = VB.Strings.Left(sg_LastLine, sg_LastLine.Length - 2)
                    ElseIf sg_LastLine.EndsWith(VB.Constants.vbCrLf) Then
                        pm_OpenedInputFile.OpenedStreamLastLines = VB.Strings.Left(sg_LastLine, sg_LastLine.Length - 1)
                    Else
                        pm_OpenedInputFile.OpenedStreamLastLines = sg_LastLine
                    End If
                End If
                Return sg_ReadLine '----------
            ElseIf pm_OpenedInputFile.OpenedStreamLastLines.Length > 0 Then
                Dim sg_LastLines As String = pm_OpenedInputFile.OpenedStreamLastLines
                Dim wk_Pos11 As Integer = VB.Strings.InStr(sg_LastLines, VB.Constants.vbCrLf)
                Dim wk_Pos22 As Integer = VB.Strings.InStr(sg_LastLines, VB.Constants.vbLf)
                If wk_Pos11 > 0 AndAlso wk_Pos11 < wk_Pos22 Then
                    pm_OpenedInputFile.OpenedStreamLastLines = VB.Strings.Mid(sg_LastLines, wk_Pos11 + 2)
                    Return VB.Strings.Left(sg_LastLines, wk_Pos11 - 1) '----------
                ElseIf wk_Pos22 > 0 Then
                    pm_OpenedInputFile.OpenedStreamLastLines = VB.Strings.Mid(sg_LastLines, wk_Pos22 + 1)
                    Return VB.Strings.Left(sg_LastLines, wk_Pos22 - 1) '----------
                Else
                    pm_OpenedInputFile.OpenedStreamLastLines = String.Empty
                    Return sg_LastLines '----------
                End If
            Else
                Throw New WorkFrameException("｢あり得ない状態 (LineInput)｣")
            End If
        End Function 'LineInput

        Private Overloads Shared Sub PrintLine(ByVal pm_OpenedOutputFile As OpenedOutputFile, ByVal pm_String As String)
            pm_OpenedOutputFile.OpenedStreamWriter.WriteLine(pm_String)
        End Sub 'PrintLine

#End Region 'TextDBMS が使用する Text File System

#Region "Text File System の内部クラス OpenedInputFile とこれに関係するメンバ"

        '========== Private Class OpenedInputFile ==============
        '#J#1/ static class OpenedInputFile
        Private Class OpenedInputFile
            '
            'OpenedInputFile (oiF) クラス
            'クラスといっても､ データの型の宣言のようなもの。
            '

            Friend nextMyObj As OpenedInputFile = Nothing
            Private ReadOnly OpenedInputFileName As String '= pm_OpenedInputFileName
            Friend ReadOnly OpenedInputFileMode As System.IO.FileMode '= pm_OpenedInputFileMode
            Friend OpenedStream As System.IO.Stream
            Friend OpenedStreamReader As System.IO.StreamReader
            Friend OpenedStreamLastLines As String = Nothing

            '========== Constructor ==========
            Friend Sub New(ByVal pm_OpenedInputFileName As String, ByVal pm_OpenedInputFileMode As System.IO.FileMode)
                Me.OpenedInputFileName = pm_OpenedInputFileName
                Me.OpenedInputFileMode = pm_OpenedInputFileMode
            End Sub 'New

        End Class 'OpenedInputFile

#End Region 'Text File System の内部クラス OpenedInputFile とこれに関係するメンバ

#Region "Text File System の内部クラス OpenedOutputFile とこれに関係するメンバ"

        '========== Private Class OpenedOutputFile ==============
        '#J#1/ static class OpenedOutputFile
        Private Class OpenedOutputFile
            '
            'OpenedOutputFile (ooF) クラス
            'クラスといっても､ データの型の宣言のようなもの。
            '

            Friend nextMyObj As OpenedOutputFile = Nothing
            Private ReadOnly OpenedOutputFileName As String '= pm_OpenedOutputFileName
            Friend ReadOnly OpenedOutputFileMode As System.IO.FileMode '= pm_OpenedOutputFileMode
            Friend OpenedStream As System.IO.Stream
            Friend OpenedStreamWriter As System.IO.StreamWriter

            '========== Constructor ==========
            Friend Sub New(ByVal pm_OpenedOutputFileName As String, ByVal pm_OpenedOutputFileMode As System.IO.FileMode)
                Me.OpenedOutputFileName = pm_OpenedOutputFileName
                Me.OpenedOutputFileMode = pm_OpenedOutputFileMode
            End Sub 'New

        End Class 'OpenedOutputFile

#End Region 'Text File System の内部クラス OpenedOutputFile とこれに関係するメンバ

#Region "TextDBMS の内部クラス RowData とこれに関係するメンバ"

        Private RowFirst As RowData
        Private RowLast As RowData
        '
        Public RecordCount As Integer 'Record の総数

        Private Sub ClearRow()
            RowFirst = Nothing
            RowLast = Nothing
            '
            RecordCount = 0
        End Sub 'ClearRow

        'ConnectHelper, InsertRecord 用の Helper メソッド (Private)。
        Private Sub AppendRowHelper(ByVal pm_row As RowData)
            If RowFirst Is Nothing Then
                RowFirst = pm_row
            Else
                pm_row.prevRow = RowLast
                RowLast.nextRow = pm_row
            End If
            RowLast = pm_row
            '
            RecordCount += 1
        End Sub 'AppendRowHelper

        'InsertRecord 用の Helper メソッド (Private)。
        Private Sub InsertPrevRowHelper(ByVal pm_row As RowData, ByVal pm_NewRow As RowData)
            If pm_row.prevRow Is Nothing Then
                '先頭エントリの前に挿入する場合
                RowFirst = pm_NewRow
            Else
                '先頭以外のエントリの前に挿入する場合
                pm_row.prevRow.nextRow = pm_NewRow
                pm_NewRow.prevRow = pm_row.prevRow
            End If
            pm_NewRow.nextRow = pm_row
            pm_row.prevRow = pm_NewRow
            '
            RecordCount += 1
        End Sub 'InsertPrevRowHelper

        'DeleteRecord 用の Helper メソッド (Private)。
        Private Sub DeleteRowHelper(ByVal pm_row As RowData)
            If pm_row.nextRow Is Nothing Then
                '最終エントリ
                If pm_row.prevRow Is Nothing Then
                    'エントリが一つだけある場合
                    RowFirst = Nothing
                Else
                    '複数のエントリがある場合
                    RowLast.prevRow.nextRow = Nothing
                End If
                RowLast = pm_row.prevRow
            Else
                '最終エントリではない
                If pm_row.prevRow Is Nothing Then
                    '先頭エントリ
                    RowFirst = pm_row.nextRow
                Else
                    '先頭エントリでも最終エントリでもない
                    pm_row.prevRow.nextRow = pm_row.nextRow
                End If
                pm_row.nextRow.prevRow = pm_row.prevRow
            End If
            '
            RecordCount -= 1
        End Sub 'DeleteRowHelper

        '========== Private Class RowData ==========
        Private Class RowData
            '
            'RowData (wk_row) クラス
            'クラスといっても､ データの型の宣言のようなもの。
            '

            Friend nextRow As RowData = Nothing
            Friend prevRow As RowData = Nothing
            Friend RowElements As String 'Row の構成要素群

            '========== Constructor ==========
            Friend Sub New(ByVal pm_RowElements As String)
                Me.nextRow = Nothing
                Me.RowElements = pm_RowElements
            End Sub 'New

        End Class 'RowData

#End Region 'TextDBMS の内部クラス RowData とこれに関係するメンバ

#End If 'ワンピースまたはセントラルピース用 (つまりローカルピースを除く)

    End Class 'TextDBMS
End Namespace 'AppliFrame.DBMS
